// sdhash_threads.cc
// author: candice quates
// threading code for hashing individual files 
// and for hashing stdin.

#include "../sdbf/sdbf_class.h"
#include "../sdbf/sdbf_defines.h"
#include "sdhash.h"

#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>

#include <boost/thread/thread.hpp>

using namespace std;

// Global parameters

extern sdbf_parameters_t sdbf_sys;


// NOT in sdbf class
/** 
    the actual processing task for the file-list hashing without block mode
*/
void 
*thread_sdbf_hashfile( void *task_param) {
    filehash_task_t *task = (filehash_task_t *)task_param;
    struct stat file_stat;
    int32_t chunks = 0,csize = 0;
    ifstream *is = new ifstream();

    int i;
    for( i=task->tid; i<task->file_count; i+=task->tcount) {
        if (stat(task->filenames[i],&file_stat))
            continue;
        is->open(task->filenames[i], ios::binary);
        try {
            class sdbf *sdbfm = new sdbf(task->filenames[i],is,0,file_stat.st_size);
            if (!task->addset) {
                cout << sdbfm;
                delete sdbfm;            
            } else 
                task->addset->add(sdbfm);
            
        } catch (int e) {
            if (e==-2)
               exit(-2);
            if (e==-3 && sdbf_sys.warnings) {
                    fprintf(stderr,"Input file too small for processing: %s\n",task->filenames[i]);
            }
       }
       is->close();
    }
	delete is;
}

sdbf_set
*sdbf_hash_stdin( ) {
    sdbf_set *myset = new sdbf_set();
    const char *default_name= "stdin";
    char *stream;
    int32_t j = 0;
    if (sdbf_sys.filename == NULL) {
        stream = (char*)default_name;
    } else {
        stream = sdbf_sys.filename;
    }
    
    char *readbuf = (char *)alloc_check(ALLOC_ONLY,sizeof(char)*sdbf_sys.segment_size, "sdbf_hash_stdin","read buffer",ERROR_EXIT);
    if (readbuf!=NULL) {
        int done = 0;
        while (!done) {
            std::stringstream namestr;
            namestr << stream;
            memset(readbuf,0,sdbf_sys.segment_size);
            size_t sz = fread(readbuf,1,sdbf_sys.segment_size,stdin);
            if (sz != sdbf_sys.segment_size) 
                done = 1;
            if (j >0) {
                namestr.fill('0');
                namestr << "." << setw(4) << j*sdbf_sys.segment_size/MB << "M" ;
			}
            string myname = namestr.str();
            char *fname = (char*)alloc_check(ALLOC_ONLY,myname.length()+1, "sdbf_hash_stdin", "generated filename",ERROR_EXIT);
            strncpy(fname,myname.c_str(),myname.length()+1);
            if (sdbf_sys.verbose) 
                fprintf(stderr,"....."); // progress bar hook
            try {
                class sdbf *sdbfm = new sdbf(fname,readbuf,sdbf_sys.dd_block_size*KB,sz);
                //myset->insert(sdbfm);
                myset->add(sdbfm);
            } catch (int e) {
                if (e==-2)
                   exit(-2);
            } 
            j++;
        }
    }
    if (sdbf_sys.verbose) 
        fprintf(stderr,"done!\n"); // progress bar.  
    return myset;
}

/**
 * Compute SD for a list of files & add them to a new set.
 * Not block-wise.
 \returns sdbf_set with the hashes in it
 */
void
sdbf_hash_files( char **filenames, uint32_t file_count, int32_t thread_cnt,sdbf_set *addto ) {
    int32_t i, t, j;
    //sdbf_set *myset = new sdbf_set();
    struct stat file_stat;
    int32_t chunks = 0,csize = 0;
    ifstream *is = new ifstream();

    // Sequential implementation
    if( thread_cnt == 1) {
        for( i=0; i<file_count; i++) {
            if (stat(filenames[i],&file_stat))
                continue;
            is->open(filenames[i], ios::binary);
            try {
                class sdbf *sdbfm = new sdbf(filenames[i],is,0,file_stat.st_size);
                //myset->insert(sdbfm);
                if (!addto) {
                            cout << sdbfm ;
                    delete sdbfm;
                } else 
                    addto->add(sdbfm);
            } catch (int e) {
                if (e==-2)
                   exit(-2);
                if (e==-3 && sdbf_sys.warnings) {
                    fprintf(stderr,"Input file too small for processing: %s\n",filenames[i]);
                }
            }
            is->close();
        }
    // Threaded implementation
    } else {
		boost::thread *thread_pooll[MAX_THREADS];
        filehash_task_t *tasks = (filehash_task_t *) alloc_check( ALLOC_ZERO, thread_cnt*sizeof( filehash_task_t), "sdbf_hash_files", "tasks", ERROR_EXIT);
        for( t=0; t<thread_cnt; t++) {
            tasks[t].tid = t;
            tasks[t].tcount = thread_cnt;
            tasks[t].filenames = filenames;
            tasks[t].file_count = file_count;
            tasks[t].addset = addto;
			thread_pooll[t] = new boost::thread(thread_sdbf_hashfile,tasks+t);
        }
        for( t=0; t<thread_cnt; t++) {
            thread_pooll[t]->join();
        }
        for( t=0; t<thread_cnt; t++) {
            delete thread_pooll[t];
        }
		free(tasks);
    // End threading
    }
    delete is;
}

