// This autogenerated skeleton file illustrates how to build a server.
// You should copy it to another filename to avoid overwriting it.

#include "sdhashsrv.h"
#include <protocol/TBinaryProtocol.h>
#include <concurrency/ThreadManager.h>
#include <concurrency/PosixThreadFactory.h>
#include <server/TSimpleServer.h>
#include <server/TThreadPoolServer.h>
#include <server/TThreadedServer.h>
#include <transport/TServerSocket.h>
#include <transport/TBufferTransports.h>

using namespace ::apache::thrift;
using namespace ::apache::thrift::protocol;
using namespace ::apache::thrift::transport;
using namespace ::apache::thrift::server;
using namespace ::apache::thrift::concurrency;

using boost::shared_ptr;

using namespace  ::sdhash;


#include <iostream>
#include <fstream>

#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_int_distribution.hpp>
#include <boost/lexical_cast.hpp>

#include "sdbf_class.h"
#include "sdbf_defines.h"
#include "sdbf_set.h"
#include "view_set.h"
#include "sdhash-srv.h"
#include "set_list.h"

/** 
    Handler class for sdhash-srv 
    Contains implmementation of service calls
*/
class sdhashsrvHandler : virtual public sdhashsrvIf {
 public:

	/** Handler initialization, starts up server
		\param thread_count number of threads to use for processing
		\param home directory to search for hashsets
	*/
  sdhashsrvHandler(uint32_t thread_count, std::string *home) {
    sdbf::config = new sdbf_conf(thread_count, 0, _MAX_ELEM_COUNT, _MAX_ELEM_COUNT_DD);
    char **mylist;
    uint32_t i,j;
    struct dirent **entries;
    uint32_t count;
    count=scandir(home->c_str(),&entries,0,0);
    j = 0;
    int dirfail=chdir(home->c_str());
    set_list_init();
	rng_init();
    sdbf_set *tmp_set;
    fprintf(stderr,"Searching home directory %s for hashsets.\n", home->c_str());
    for (i= 0; i< count; i++) {
        if (entries[i]->d_type == DT_REG) {
            try {
                fprintf(stderr,"-> Loading file %s ..... ",entries[i]->d_name);
                tmp_set=new sdbf_set(entries[i]->d_name);
                if (!tmp_set->empty()) {
                    add_set(tmp_set,entries[i]->d_name);
                    j++;
                    fprintf(stderr,"OK\n");
                } else {
                    fprintf(stderr,"File empty\n",entries[i]->d_name);
                }
            } catch (int e) {
                if (e==-1) 
                    fprintf(stderr,"File not accessible\n");
                if (e==-2) 
                    fprintf(stderr,"File format invalid\n");
                if (e==-3) 
                    fprintf(stderr,"File too small\n");

            }
        }
    }
    fprintf(stderr, "sdhash-srv loaded %d sets.\n",j );
  }

  /** 
 	ping()	
  */
  void ping() {
    // Your implementation goes here
    printf("ping\n");
  }

  /** 
	retrieves listing of sets
	\param _return string of listing	
	\param json use json output if true
  */
  void setList(std::string& _return, const bool json) {
    int i; 
    int defchar='K';  
    uint64_t sizetmp=0;
    cout << "setList request" << endl;
    std::stringstream out;
	if (json) 
		out << "[";
    for (i=0; i < get_set_count(); i++)  {
        sizetmp = get_set(i)->input_size();
        if (sizetmp > GB)  {
            sizetmp = sizetmp/ GB;
            defchar = 'G';
        } else if (sizetmp > MB ) {
            sizetmp = sizetmp / MB;
            defchar = 'M';
        } else  {
            sizetmp = sizetmp / KB;
            defchar = 'K';
        }
		if (json) {
			if (i!=0)
				out << ",";
			out << "{\""<< i << "\":{\"" ;
			out << get_set_name(i) <<"\":\"" ;
			out << sizetmp << (char) defchar << "\"}}" ;
		} else {
			out << i << " " ;
			out << get_set_name(i) << " " ;
			out << sizetmp << (char) defchar << "\n" ;
		}
    }
	if (json) 
		out << "]";
    if (get_set_count() == 0) {
		if (json)
			_return="[\"sdhash-srv ready, no sets loaded\"]";
		else 
			_return="sdhash-srv ready, no sets loaded\n";
    } else 
        _return=out.str();
  }

  /** 
	retrieves listing of set contents and sizes
	\param _return string of listing
	\param num1 number of set
  */
  void displaySet(std::string& _return, const int32_t num1) {
    std::stringstream out;
    uint64_t sizetmp=0;    
    int defchar='K';
    if ((num1 > get_set_count()) || num1 < 0 ) {
        std::cout << "displaySet request for " << num1 << " set not found." << std::endl;
        _return="Error: not found\n";
    } else {
        std::cout << "displaySet request for " << num1 << " " << get_set_name(num1) << std::endl;
        sdbf_set *target=get_set(num1);
        for (int i=0; i < target->size(); i++) {
            out << target->at(i)->name() << " ";
            sizetmp=target->at(i)->input_size();     
            if (sizetmp > GB)  {
                sizetmp = sizetmp/ GB;
                defchar = 'G';
            } else if (sizetmp > MB ) {
                sizetmp = sizetmp / MB;
                defchar = 'M';
            } else  {
                sizetmp = sizetmp / KB;
                defchar = 'K';
            }
            out << sizetmp << (char) defchar << "\n" ;
        }
        _return=out.str();
    }
  }

  /** 
	retrieves set hash contents 
	\param _return string of displayable hashes
	\param num1 number of set
  */
  void displayContents(std::string& _return, const int32_t num1) {
    std::stringstream out;
    if ((num1 > get_set_count()) || num1 < 0 ) {
        _return="Error: not found\n";
        std::cout << "displayContents request for " << num1 << " set not found." << std::endl;
    } else {
        std::cout << "displayContents request for " << num1 << " " << get_set_name(num1) << std::endl;
        out << get_set(num1);
        _return=out.str();
    }
  }

  /** 
	compares all hashes in passed set, associates them with ID when finished
	\param num1 number of set
	\param threshold output confidence threshold
	\param resultID ID to store results under
  */
  void compareAll(const int32_t num1, const int32_t threshold, const int32_t resultID) {
	std::string result_str;
    if (num1 < get_set_count() || num1 < 0 ) {
		add_request(resultID,"compareAll "+(string)get_set_name(num1));
        std::cout << "compareAll begin request for " << num1 << " " << get_set_name(num1) <<  std::endl;
        result_str=get_set(num1)->compare_all(threshold);
		add_result(resultID,result_str);
    } else {
        std::cout << "compareAll begin request for " << num1 << " not found" << std::endl;
    }
    std::cout << "compareAll request ends for " << num1 << std::endl;
  }

  /** 
	compares two passed sets, associates them with ID when finished
	\param num1 number of set
	\param num2 number of set
	\param threshold output confidence threshold
	\param sample count of filters to sample in query
	\param resultID ID to store results under
  */
  void compareTwo(const int32_t num1, const int32_t num2, const int32_t threshold, const int32_t sample, const int32_t resultID) {
	std::string result_str;
    if ((num1 < get_set_count() || num1 < 0 ) &&  (num2 < get_set_count() || num2 < 0 ))  {
		add_request(resultID,"compareTwo "+(string)get_set_name(num1)+" "
			+(string)get_set_name(num2));
        std::cout << "compareTwo begin request for " << num1 << " " << get_set_name(num1) ;
        std::cout << " "<< num2 << " " << get_set_name(num2) <<  std::endl;
        result_str=get_set(num1)->compare_to(get_set(num2),threshold,sample);
		add_result(resultID,result_str);
    } else {
        std::cout << "compareTwo begin request for " << num1 ;
        std::cout << ", "<< num2 << " sets not found" << std::endl;
    }
    std::cout << "compareTwo request ends for " << num1 ;
    std::cout << " "<< num2 <<  std::endl;
  }

  /** 
	loads a hashset from a file.  
	\param filename filename containing hash set
	\returns int32_t id, < 0 on failure
  */
  int32_t loadSet(const std::string& filename) {
    sdbf_set *tmp_set;
    char *setname;
    setname=(char*)malloc(filename.length()+1);
    strncpy(setname,filename.c_str(),filename.length()+1);
    int32_t setid=-1;
    int fail=0;
    std::cout << "loadSet begin request to load " << filename << std::endl;
    try {
        tmp_set=new sdbf_set(filename.c_str());
        if (!tmp_set->empty()) {
            add_set(tmp_set,setname);
            setid=get_set_count()-1;
        } else {
            fprintf(stderr,"File not found or empty: %s\n",filename.c_str());
            fail=1;
        }
    } catch (int e) {
        if (e==-1) 
            fprintf(stderr,"File not accessible: %s\n",filename.c_str());
        if (e==-2) 
            fprintf(stderr,"File format invalid: %s\n",filename.c_str());
        if (e==-3) 
            fprintf(stderr,"File too short: %s\n",filename.c_str());
        fail=1;
    }
    if (!fail)
        std::cout << "loadSet request successful for " << filename << std::endl;
    else
        std::cout << "loadSet request failed for " << filename << std::endl;
    return setid;
  }

  int32_t saveSet(const int32_t num1, const std::string& filename) {
    std::cout << "saveSet begin request for "<< num1 << " " << filename << std::endl;
    if (num1 >= 0 && num1 < get_set_count()) {
        std::filebuf fb;
        fb.open (filename,ios::out);
        if (fb.is_open()) {
            std::ostream os(&fb);
            os << get_set(num1);
            fb.close();
            std::cout << "saveSet request succeeded "<< num1 << " " << filename << std::endl;
        } else {
            fprintf(stderr,"saveSet cannot write to file %s\n",filename.c_str());
            return -1;
        }
    } else  {
        fprintf(stderr,"saveSet set number not valid %d\n",num1);
        return -1;
    }
    return 0;
  }
  void hashList(const std::string& filename, const int32_t blocksize) {
    sdbf_set *tmp;
    std::cout << "hashList begin request for "<< filename << std::endl;
    tmp=hash_list((char*)filename.c_str(),blocksize);
    if (!tmp->empty()) {
        char *setname;
        setname=(char*)malloc(filename.length()+1);
        strncpy(setname,filename.c_str(),filename.length()+1);
        add_set(tmp, setname);
        std::cout << "hashList request succeeds for "<< filename << std::endl;
    } else {
        std::cout << "hashList request Failed for "<< filename << std::endl;
    }
  }

  int32_t createView(const std::string& name, const int32_t num1, const std::string& regex) {
    // Your implementation goes here
    int32_t retval= -1;
    sdbf_set *tmp;
    std::cout << "createView begin request for "<< num1 << " " << regex << std::endl;
    if (num1 >=0 && num1 < get_set_count()) {
        tmp=create_view(get_set(num1),(char*)regex.c_str());
        if (!tmp->empty()) {
            char *setname;
            setname=(char*)malloc(name.length()+1);
            strncpy(setname,name.c_str(),name.length()+1);
            add_set(tmp, setname);
            retval=get_set_count()-1;
            std::cout << "createView request found matches for "<< num1 << " " << regex << std::endl;
        } 
    }
    std::cout << "createView request ends for "<< num1 << " " << regex << std::endl;
    return retval;
  }

  int32_t findSet(const std::string& nameregex) {
    // Your implementation goes here
    printf("findSet\n");
  }


  void getHashsetName(std::string& _return, const int32_t num1) {
    // Your implementation goes here
	if (num1 > -1 && num1 < get_set_count())
		_return=get_set_name(num1);
	else 
		_return="";
    printf("getHashsetName\n");
  }
  
  void displayResult(std::string& _return, const int32_t resultID) {
	_return=get_request(resultID)+get_result(resultID);
    cout << "displayResult " << resultID << endl;
  }

  void displayResultsList(std::string& _return, const std::string& user, const bool json) {
	std::stringstream resultlist;
	multimap<string,int32_t>::iterator it;
	if (json)
		resultlist << "[";
    for (it=users.equal_range(user).first; it!=users.equal_range(user).second; ++it) {
		if (json) {
			if (it!=users.equal_range(user).first)
				resultlist << ",";
			resultlist << "{\"" << (*it).second <<"\":\"";
			resultlist << get_request((*it).second) << "\"}";
		} else {
			resultlist << (*it).second << " " ;
			resultlist << get_request((*it).second) << endl;
		}
	}
	if (json)
		resultlist << "]";
	_return=resultlist.str();
    cout << "displayResultsList " << user << endl;
  }

  int32_t createResultID(const std::string& user) {
    cout << "createResultID " << user << endl;
	return add_resultID(user);
  }

  void shutdown() {
    printf("sdhash-srv shutdown requested by client. Shutting down.\n");
    server_->stop();
  }

  void setServer(boost::shared_ptr<TServer> server) {
    server_ = server;
  }

private:

    boost::shared_ptr<TServer> server_;

	multimap<string,int32_t> users;
	map<int32_t,string> requests;
	map<int32_t,string> results;
	map<int32_t,time_t> starttime;
	map<int32_t,time_t> endtime;

	boost::random::mt19937 gen;

	void
	rng_init() {
		gen.seed(static_cast<unsigned int>(time(0)));
	}

	void
	add_result(int32_t resultID, string res) {
		results.insert(pair<int32_t,string>(resultID,res));
		endtime.insert(pair<int32_t,time_t>(resultID,time(0)));
	}

	void
	add_request(int32_t resultID, string req) {
		requests.insert(pair<int32_t,string>(resultID,req));
		starttime.insert(pair<int32_t,time_t>(resultID,time(0)));
	}

	string	
	get_request(int32_t resultID) {
		string result="";
		std::map<int32_t,string>::iterator reqfound = requests.find(resultID);
		if (reqfound!=requests.end() ) {
			string idstr=boost::lexical_cast<string>(resultID);
			result= "query "+idstr+" "+reqfound->second;
	    }
		return result;
	}

	int32_t
	add_resultID(string user) {
		boost::random::uniform_int_distribution<> dist(1,131070);
		int32_t id = dist(gen);
		users.insert(pair<string,int32_t>(user,id));
		return id;
	}

	string
	get_result(int32_t resultID) {
		string result="";
		std::map<int32_t,string>::iterator resfound = results.find(resultID);
		std::map<int32_t,time_t>::iterator startfound = starttime.find(resultID);
		std::map<int32_t,time_t>::iterator endfound = endtime.find(resultID);
		if (resfound!=results.end() && startfound != starttime.end() && endfound != endtime.end() )  {
			time_t querytime = (endfound->second - startfound->second);
			string timestr=boost::lexical_cast<string>(querytime);
			string idstr=boost::lexical_cast<string>(resultID);
			result= resfound->second+"query "+idstr+" time "+timestr + " seconds\n";
		}
		return result;
	}

};

int main(int argc, char **argv) {
    sdbf_parameters_t* conf=NULL;

    conf=read_args(argc,argv,0);    
    if (conf==NULL) 
        return 0;
    fprintf(stderr, "sdhash-srv initializing service...\n");
    boost::shared_ptr<sdhashsrvHandler> handler(new sdhashsrvHandler(conf->thread_cnt,conf->home));
    boost::shared_ptr<TProcessor> processor(new sdhashsrvProcessor(handler));
    boost::shared_ptr<TServerTransport> serverTransport(new TServerSocket(conf->port));
    boost::shared_ptr<TTransportFactory> transportFactory(new TBufferedTransportFactory());
    boost::shared_ptr<TProtocolFactory> protocolFactory(new TBinaryProtocolFactory());


    boost::shared_ptr<ThreadManager> threadManager = ThreadManager::newSimpleThreadManager(conf->maxconn);
    boost::shared_ptr<PosixThreadFactory> threadFactory = boost::shared_ptr<PosixThreadFactory>(new PosixThreadFactory());

    threadManager->threadFactory(threadFactory);
    threadManager->start();

    boost::shared_ptr<TServer> server = boost::shared_ptr<TServer>
            (new TThreadPoolServer(processor, serverTransport, transportFactory, protocolFactory, threadManager));
    handler->setServer(server);
    fprintf(stderr,"sdhash-srv listening on port %d, with %d worker threads.\n",conf->port, conf->maxconn);
    server->serve();
    return 0;
}

